package nq.utils.task {

    import flash.errors.IllegalOperationError;
    import flash.events.Event;
    import flash.utils.Dictionary;

    import nq.core.IDestroyable;
    import nq.utils.EnterFrameBroadcaster;
    import nq.utils.date.IDateProvider;

    /**
     * Планировщик заданий
     */
    public class TaskScheduler implements IDestroyable {

        private var _dateProvider:IDateProvider;
        private var _hash:Dictionary;
        private var _head:Node;

        /**
         * @param dateProvider  Объект, ответственный за предоставление текущего времени
         */
        public function TaskScheduler(dateProvider:IDateProvider) {

            super();

            if (!dateProvider) throw new ArgumentError("Параметр не должен быть нулевым!");

            _dateProvider = dateProvider;
            _hash = new Dictionary();

        }

        /**
         * Запланировать запуск задания
         *
         * @param task      Планируемое задание
         * @param date      Дата выполнения задания (в миллисекундах)
         */
        public function schedule(task:ITask, date:Number = 0):void {

            if(!task) throw new ArgumentError("Параметр не должен быть нулевым!");
            if (isNaN(date) || date == Number.NEGATIVE_INFINITY || date == Number.POSITIVE_INFINITY) throw new ArgumentError("Дата должна быть представлена числом!");
            if (task in _hash) throw new IllegalOperationError("Задание уже запланировано!");

            this.insertNode(new Node(task, date));

        }

        /**
         * Отложить запуск задания
         *
         * @param task  Планируемое задание
         * @param delay Время в миллисекундуах, через которое нужно запустить задание
         */
        public function delay(task:ITask, delay:Number = 0):void {

            if(!task) throw new ArgumentError("Параметр не должен быть нулевым!");
            if (isNaN(delay) || delay == Number.NEGATIVE_INFINITY || delay == Number.POSITIVE_INFINITY) throw new ArgumentError("Дата должна быть представлена числом!");
            if (task in _hash) throw new IllegalOperationError("Задание уже запланировано!");

            this.insertNode(new Node(task, _dateProvider.getTime() + delay));

        }

        /**
         * Отменить запланированное задание
         *
         * @param task  Запланированное задание
         */
        public function cancelTask(task:ITask):void {

            if(!task) throw new ArgumentError("Параметр не должен быть нулевым!");
            if (!(task in _hash)) throw new IllegalOperationError("Задание не было запланировано!");

            var node:Node = _hash[task] as Node;
            delete _hash[task];

            this.destroyNode(node);

            if (!_head) EnterFrameBroadcaster.instance.removeEventListener(Event.ENTER_FRAME, onFrame);
        }

        /**
         * Было ли задание запланировано
         *
         * @param task  Запланированное задание
         */
        public function isScheduled(task:ITask):Boolean {

            if(!task) throw new ArgumentError("Параметр не должен быть нулевым!");
            return (task in _hash);
        }

        private function insertNode(node:Node):void {

            // добавляем в хэш
            _hash[node.task] = node;

            // добавляем в список (в начале списка стоят задания с меньшей датой)
            if (!_head)
            {

                _head = node;
                
                EnterFrameBroadcaster.instance.addEventListener(Event.ENTER_FRAME, onFrame);

            }
            else
            {
                
                if (_head.date > node.date) { // ставим node первым в очереди

                    node.next = _head;
                    _head.prev = node;

                    _head = node;

                }else { // ставим node в середину очереди

                    var next:Node = _head;

                    while (next.next && next.next.date < node.date) next = next.next;

                    if (next.next) { // вставляем между next и next.next

                        node.next = next.next;
                        node.prev = next;

                        next.next.prev = node;
                        next.next = node;

                    }else { // вставляем в конец очереди

                        next.next = node;
                        node.prev = next;

                    }

                }

            }
        }

        private function destroyNode(node:Node):void {

            node.task = null;

            if (_head == node) _head = node.next;

            if (node.next) node.next.prev = node.prev;
            if (node.prev) node.prev.next = node.next;

            node.next = node.prev = null;
            
        }

        private function onFrame(event:Event):void {

            var date:Number = _dateProvider.getTime();

            var task:ITask;


			while (_head && (_head.date <= date)) {

                task = _head.task;

                delete _hash[_head.task];

                this.destroyNode(_head);


                task.start();
            }

            if (!_head) EnterFrameBroadcaster.instance.removeEventListener(Event.ENTER_FRAME, onFrame);

        }

        
        /* INTERFACE nq.core.IDestroyable */

        /**
         * Деструктор объекта
         */
        public function destroy():void {

            while(_head) this.destroyNode(_head);
            
            _head = null;
            _hash = null;

            EnterFrameBroadcaster.instance.removeEventListener(Event.ENTER_FRAME, onFrame);
            
        }

    }

}


import nq.utils.task.ITask;

internal final class Node {

    public var next:Node;
    public var prev:Node;
    public var task:ITask;
    public var date:Number;

    /**
     * @param task      Планируемое задание
     * @param date      Дата выполнения задания
     */
    public function Node(task:ITask, date:Number) {

        super();

        this.task = task;
        this.date = date;

    }

}